#include "GaugeCluster.h"



// +-------------------+
// | Class Constructor |
// +-------------------+
Gauge::Gauge(Stream * virtualPort, int intBTN, void * parentCluster) : Display(virtualPort) {

	Display.SSTimeout(0);
	Display.txt_FGcolour(WHITE);
	Display.gfx_MoveTo(38, 60);        
	Display.putstr("RetroJDM.com");


	while (!Display.media_Init()) {
		Display.gfx_Cls();
		Display.gfx_MoveTo(3, 60);
		Display.putstr("Please insert SD card.");
		delay(300);
	}
	
	// Initialise some important settings.
	cluster					= parentCluster;
	mode					= MODE_SCREENS;
	minLastDrawn			= -1;
	screenInitialised		= false;

	// Button
	buttonPin				= intBTN;
	buttonWas				= LOW;
	buttonIs				= LOW;
	buttonMillis			= 0;
	buttonHeld				= false; // Has the button been held?

	// Attributes
	// Menu
	itemCount				= 0;
	parent					= 0;
	offset					= 0;
	cursor					= 0;

	// Font
	font					= fontM;
	cursorX					= 0;
	cursorY					= 0;


	pinMode(buttonPin, INPUT_PULLUP);
}



void Gauge::begin(void) {

	screenInitialised = false;

	switch (mode) {
	
	case MODE_SCREENS:
		setScreen(screen);
		break;

	case MODE_MENU:
		setMenu(parent);
		break;
	}
}



void Gauge::loop(void) {

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

	// Do we need to refresh a screen?
	if (mode == MODE_SCREENS) {
		switch(screen) {

		case SCREEN_CLOCK:
			if (
				(ptrCluster->fix != ptrCluster->fixWas) ||
				((!ptrCluster->fix) && (ptrCluster->frame != ptrCluster->frameWas)) ||
				(ptrCluster->minute != ptrCluster->minuteWas)
			) refresh();
			break;

		case SCREEN_SPEEDO_MPH:
		case SCREEN_SPEEDO_KMPH:
		case SCREEN_SPEEDO_KNOTS:
			if (
				(ptrCluster->fix != ptrCluster->fixWas) ||
				((!ptrCluster->fix) && (ptrCluster->frame != ptrCluster->frameWas)) ||
				(ptrCluster->knots != ptrCluster->knotsWere)
			) refresh();
			break;

		case SCREEN_TACHO:
			if (ptrCluster->rpm != ptrCluster->rpmWas) refresh();
			break;

		case SCREEN_HEADING:
			if (
				(ptrCluster->fix != ptrCluster->fixWas) ||
				((!ptrCluster->fix) && (ptrCluster->frame != ptrCluster->frameWas)) ||
				(ptrCluster->heading != ptrCluster->headingWas)
			) refresh();
			break;

		case SCREEN_GPS_HOME:
		case SCREEN_GPS_LOCATION:
			if (
				(ptrCluster->fix != ptrCluster->fixWas) ||
				((!ptrCluster->fix) && (ptrCluster->frame != ptrCluster->frameWas)) ||
				(ptrCluster->latitude != ptrCluster->latitudeWas) ||
				(ptrCluster->latitudeNS != ptrCluster->latitudeNSWas) ||
				(ptrCluster->longitude != ptrCluster->longitudeWas) ||
				(ptrCluster->longitudeEW != ptrCluster->longitudeEWWas)
			) refresh();
			break;

		}
	}


	// Buttons
	buttonWas	= buttonIs;
	buttonIs	= digitalRead(buttonPin) == LOW;

	// Is the button down?
	if (buttonIs) {

		// Has the button NOT yet been flagged as "Held"?
		if (!buttonHeld) {

			// Has the button just changed to "on"?
			// (We know the button is down. If it wasn't down just before, then it must have been pressed)
			if (!buttonWas) {
				buttonHeld		= false;
				buttonMillis	= millis();
			}

			// Has the button been held long enough?
			if (millis() > buttonMillis + GAUGE_BUTTON_HOLD) {

				// Flag it as held. Trigger the OK event.
				buttonHeld = true;
				ok();
			}
		}
	}

	// Button just changed to "off".
	if (!buttonIs && buttonWas) {		

		if (buttonHeld) {
			buttonHeld = false;
		}
		else {
			// Release Event
			select();
		}
	}
}



// Button presses
// --------------------------------------------------------------------------------
void Gauge::select(void) {

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

	word intScreenWas = screen;

	switch (mode) {

	case MODE_SCREENS:
		
		// Go to the next gauge screen.
		// We cycle through them until we either:
		//  - find one that's enabled, or
		//  - loop back to the one we started on
		do {
			screen++;
			if (screen >= SCREENS) screen = 0;
		} while (!(screens[screen]) && (screen != intScreenWas));
		

		ptrCluster->saveSettings();

		screenInitialised = false;
		setScreen(screen);
		break;

	case MODE_MENU:

		// go to the next menu item.
		cursor++;
		if (cursor >= itemCount) cursor = 0;

		// Scroll to ensure visible
		if (offset > cursor) offset = cursor;
		if (offset < cursor - ITEMS_PER_SCREEN + 1) offset = cursor - ITEMS_PER_SCREEN + 1;

		drawMenu();
		break;

	case MODE_DIALOG:

		// Go back to the menu.
		mode = MODE_MENU;
		drawMenu();
		break;
	}
}



void Gauge::ok(void) {
	
	MenuItemType objItem = items[cursor];

	// If we're in menu mode
	switch (mode) {

	case MODE_SCREENS:

		// We're already in Screen mode. So pressing a button put's it in Menu mode.
		mode = MODE_MENU;
		screenInitialised = false;
		setMenu(0); // Start the main menu.
		break;


	case MODE_MENU:

		// What kind of item are we OK'ing ?
		switch (objItem.type) {

		case TYPE_ITEM:
			
			// Make sure the screen / menu is initialised.
			screenInitialised = false;
			
			// Execute the command
			switch (objItem.command) {
			case CMD_EXIT:		exit();						break;
			case CMD_SET_GPS:	setGPSHome();				break;
			case CMD_ABOUT:		about();					break;
			case CMD_SET_MENU:	setMenu(objItem.value);		break;
			}
			break;

		case TYPE_CHECKBOX:
			// Toggle the checkbox
			setSetting(objItem, 1 - (getSetting(objItem)));
			drawMenu();
			break;

		case TYPE_RADIOBUTTON:
			// Clear all the radio buttons in this menu
			setSetting(objItem, objItem.value);
			drawMenu();
			break;
		}

		break;

	case MODE_DIALOG:

		// Go back to the menu.
		mode = MODE_MENU;
		drawMenu();
		break;
	}
}



void Gauge::refresh(void) {

	switch (mode) {

	case MODE_SCREENS:
		drawScreen(screen);
		break;

	case MODE_MENU:
		drawMenu();
		break;
	}
}



// Fonts
// --------------------------------------------------------------------------------

int Gauge::fontWidth(char *strText) {

	// Returns the width in pixels of the string.

	int intWidth	= 0;
	int intChars	= strlen(strText);
	char ch;

	// Step through each character, adding up the widths.
	// Use ".spaceWidth" if the character is not supported.
	for (int c = 0; c < intChars; c++) {
		ch = strText[c];
		if ((ch >= font.firstChar) && (ch < font.firstChar + font.characters))
			intWidth += font.characterWidths[ch - font.firstChar];
		else intWidth += font.spaceWidth;
	}
	
	// Add all the gaps between characters.
	intWidth += font.letterSpacing * (intChars - 1);

	return intWidth;
}



void Gauge::fontWrite(char ch) {

	// Places on character at the display's GFX cursor position.
	// This function assumes the call to "Display.media_SetAdd(font.address);" has already been made.

	int intLeft, intRight;

	if ((ch >= font.firstChar) && (ch < font.firstChar + font.characters)) {
		
		// Place the character's graphic.
		Display.media_VideoFrame(cursorX, cursorY, ch - font.firstChar);
		
		// Move the cursor.
		cursorX += font.characterWidths[ch - font.firstChar];

	} else {
		// Clear for the blank character
		Display.gfx_RectangleFilled(cursorX, cursorY, cursorX + font.characterWidths[ch - font.firstChar] - 1, cursorY + font.lineHeight - 1, BLACK);
		
		// Move the cursor 
		cursorX += font.spaceWidth;
	}
	
	// Add the gap between characters.
	intLeft = cursorX;
	cursorX += font.letterSpacing;
	intRight = cursorX < DISPLAY_WIDTH ? cursorX : DISPLAY_WIDTH - 1;

	// Clear for the gap between characters
	if ((intLeft < DISPLAY_WIDTH) && (intLeft != cursorX))
		Display.gfx_RectangleFilled(intLeft, cursorY, cursorX, cursorY + font.lineHeight - 1, BLACK);
}



void Gauge::fontPrint(char *strText, byte x, byte y, byte align, boolean clearLeft, boolean clearRight) {

	// Prints a single line in the bigger font.

	int intWidth = fontWidth(strText);
	int intLeft, intRight;

	// Get the font's graphic data.
	Display.media_SetAdd((word)(font.address >> 16), (word)font.address);

	// Where to start?
	switch (align) {
	case TEXT_ALIGN_LEFT:		intLeft = x;					break;
	case TEXT_ALIGN_RIGHT:		intLeft = x - intWidth;			break;
	case TEXT_ALIGN_CENTER:		intLeft = x - (intWidth / 2);	break;
	}
	intRight = intLeft + intWidth - 1;

	cursorX = intLeft;
	cursorY = y;

	// Clear left?
	if (clearLeft) Display.gfx_RectangleFilled(0, y, intLeft - 1, y + font.lineHeight - 1, BLACK);
	
	// Place all the characters.
	for (int c = 0; c < strlen(strText); c++) fontWrite(strText[c]);

	// Clear right?
	if (clearRight) Display.gfx_RectangleFilled(intRight + 1, y, DISPLAY_WIDTH - 1, y + font.lineHeight - 1, BLACK);
}




// Screens
// --------------------------------------------------------------------------------
void Gauge::setGPSHome(void) {

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;	

	// Set home as the current GPS location.
	ptrCluster->latitudeHome	= ptrCluster->latitude;
	ptrCluster->latitudeHomeNS	= ptrCluster->latitudeNS;
	ptrCluster->longitudeHome	= ptrCluster->longitude;
	ptrCluster->longitudeHomeEW	= ptrCluster->longitudeEW;
	
	// Save it to EEPROM
	ptrCluster->saveSettings();
	
	// "GPS Home Set" screen
	Display.gfx_Cls();
	Display.gfx_MoveTo(0, 0);

	// Icon
	Display.media_SetAdd((word)(IMAGE_ICON_HOMEPIN >> 16), (word)IMAGE_ICON_HOMEPIN);
	Display.media_Image(70, 42);

	// GPS Home Set
	font = fontS_White;
	//font = colours[ptrCluster->colour].font;
	fontPrint("GPS HOME SET", DISPLAY_WIDTH / 2, 66, TEXT_ALIGN_CENTER, false, false);

	mode = MODE_DIALOG;
	
	// Show the changes on other screens if there are any.
	ptrCluster->refreshAllGauges();
}



void Gauge::about(void) {

	char strBuffer[25];

	// Get a handle on this gauge's parent cluster.
	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;
	
	Display.gfx_Cls();

	// Border for screen alignment.
	//Display.gfx_Rectangle(0, 0, DISPLAY_WIDTH - 1, DISPLAY_HEIGHT - 1, WHITE);

	// "RetroJDM.com" logo
	Display.media_SetAdd((word)(IMAGE_RETROJDM >> 16), (word)IMAGE_RETROJDM);
	Display.media_Image(0, 44);

	// "Gauge Cluster"
	// font = fontS_White;
	font = colours[ptrCluster->colour].font;
	fontPrint("GAUGE CLUSTER", DISPLAY_WIDTH / 2, 64, TEXT_ALIGN_CENTER, false, false);

	// Revision
	Display.txt_FGcolour(WHITE);
	Display.gfx_MoveTo(45, 84);
	Display.putstr((char *)GAUGE_CLUSTER_VERSION);

	mode = MODE_DIALOG;
}



void Gauge::setScreen(byte bytScreen) {
	
	screen = bytScreen;

	// If this screen is disabled, go to the next one.
	if (!screens[screen]) {
		do {
			screen++;
			if (screen >= SCREENS) screen = 0;
		} while (!(screens[screen]) && (screen != bytScreen));
	}
	mode = MODE_SCREENS;

	if (screens[screen]) drawScreen(screen);
	else Display.gfx_Cls();
}



void Gauge::drawScreen(byte bytScreen) {

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;
	
	// Redraw the entire screen if we get or lose satellite fix.
	if (ptrCluster->fix != ptrCluster->fixWas) screenInitialised = false;

	// Which screen are we drawing?
	switch (bytScreen) {

	case SCREEN_SPEEDO_MPH:
		if (ptrCluster->fix || !(ptrCluster->speedoNeedsFix)) drawSpeedo("MPH", VIDEO_SPEEDO_ANALOG_MPH, (byte)(ptrCluster->knots * MPH_PER_KNOT), VIDEO_SPEEDO_ANALOG_MPH_PER_FRAME, VIDEO_SPEEDO_ANALOG_MPH_LAST_FRAME);
		else drawSatellite("MPH");
		break;

	case SCREEN_SPEEDO_KMPH:
		if (ptrCluster->fix || !(ptrCluster->speedoNeedsFix)) drawSpeedo("KMPH", VIDEO_SPEEDO_ANALOG_KMPH, (byte)(ptrCluster->knots * KMPH_PER_KNOT), VIDEO_SPEEDO_ANALOG_KMPH_PER_FRAME, VIDEO_SPEEDO_ANALOG_KMPH_LAST_FRAME);
		else drawSatellite("KMPH");
		break;

	case SCREEN_SPEEDO_KNOTS:
		if (ptrCluster->fix || !(ptrCluster->speedoNeedsFix)) drawSpeedo("KNOTS", VIDEO_SPEEDO_ANALOG_KNOTS, (byte)(ptrCluster->knots), VIDEO_SPEEDO_ANALOG_KNOTS_PER_FRAME, VIDEO_SPEEDO_ANALOG_KNOTS_LAST_FRAME);
		else drawSatellite("KNOTS");
		break;

	case SCREEN_TACHO:
		drawTacho();
		break;

	case SCREEN_CLOCK:
		if (ptrCluster->fix || !(ptrCluster->clockNeedsFix)) drawClock();
		else drawSatellite("CLOCK");
		break;

	case SCREEN_HEADING:
		if (ptrCluster->fix || !(ptrCluster->headingNeedsFix)) drawHeading();
		else drawSatellite("HEADING");
		break;

	case SCREEN_GPS_HOME:
		if (ptrCluster->fix) drawGPSHome();
		else  drawSatellite("GPS HOME");
		break;

	case SCREEN_GPS_LOCATION:
		if (ptrCluster->fix) drawGPSLocation();
		else  drawSatellite("GPS LOCATION");
		break;
	}
}



void Gauge::drawSpeedo(char *strLabel, DWORD dwrdBackground, byte bytSpeed, byte bytSpeedPerFrame, byte bytLastFrame) {

	char strBuffer[4];
	byte bytLen;

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;	

	// Initialise
	if (!screenInitialised) {
		Display.gfx_Cls();

		switch (ptrCluster->style) {	
		case STYLE_DIGITAL:
			
			// Print the label in the small font
			font = colours[ptrCluster->colour].font;
			fontPrint(strLabel, DISPLAY_WIDTH / 2, DISPLAY_HEIGHT - font.lineHeight, TEXT_ALIGN_CENTER, false, false);

			font = fontXXXL;
			break;

		case STYLE_ANALOG:
			break;
		}
		
		screenInitialised = true;
	}


	// Draw
	switch (ptrCluster->style) {	
	case STYLE_DIGITAL:

		sprintf(strBuffer, "%d", bytSpeed);
		bytLen = strlen(strBuffer);

		// Draw the digits.
		fontPrint(strBuffer, DISPLAY_WIDTH / 2, 0, TEXT_ALIGN_CENTER, true, true);

		break;
	
	case STYLE_ANALOG:

		// Background image
		Display.media_SetAdd((word)(dwrdBackground >> 16), (word)dwrdBackground);
		Display.media_VideoFrame(0, 0, constrain(bytSpeed / bytSpeedPerFrame, 0, bytLastFrame));
		break;
	}
}



void Gauge::drawTacho(void) {

	char strBuffer[4];

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;	

	// Initialise
	if (!screenInitialised) {
		
		// Clear
		Display.gfx_Cls();

		switch (ptrCluster->style) {
		
		case STYLE_DIGITAL:

			// Print the label in the small font
			font = colours[ptrCluster->colour].font;
			fontPrint("RPM", DISPLAY_WIDTH / 2, DISPLAY_HEIGHT - font.lineHeight, TEXT_ALIGN_CENTER, false, false);
		
			break;
		
		case STYLE_ANALOG:
			break;
		}
		
		screenInitialised = true;
	}


	// Draw
	switch (ptrCluster->style) {
	
	case STYLE_DIGITAL:
		
		// Large digits
		font = fontXXL;
		sprintf(strBuffer, "%d", ptrCluster->rpm);
		fontPrint(strBuffer, 80, 20, TEXT_ALIGN_CENTER, true, true);
		break;
	
	case STYLE_ANALOG:
		
		// Background image
		Display.media_SetAdd((word)(VIDEO_TACHO_ANALOG >> 16), (word)VIDEO_TACHO_ANALOG);
		Display.media_VideoFrame(0, 0, constrain(ptrCluster->rpm / VIDEO_TACHO_ANALOG_RPM_PER_FRAME, 0, VIDEO_TACHO_ANALOG_LAST_FRAME));
		
		break;
	}
}



void Gauge::drawClock(void) {

	static const char *arrMonths[] = { "JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC" };
	
	char strTime[6]; // "hh:mm"
	char strDate[7]; // "dd-mmm"

	// Get a pointer to the parent GaugeCluster object.
	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

	// Clear
	if (!screenInitialised) {
		Display.gfx_Cls();
		screenInitialised = true;
	}


	int intHour24	= ptrCluster->hour;
	int intMinute	= ptrCluster->minute;
	int intDay		= ptrCluster->day;
	int intMonth	= ptrCluster->month;
	int intYear		= ptrCluster->year;
	int x1, y1, x2, y2;
	float rad;

	int intRemainder;

	// Adjust for the timezone
	intMinute += ptrCluster->timezone;
	
	// Roll over minute
	if ((intMinute < 0) || (intMinute > 59)) {
		intHour24		+= intMinute / 60;
		intRemainder	= abs(intMinute) % 60;
		if (intMinute < 0) intMinute = 60 - intRemainder;
		else intMinute = intRemainder;
	}
	
	// Roll over hour
	if ((intHour24 < 0) || (intHour24 > 23)) {
		intDay			+= intHour24 / 24;
		intRemainder	= abs(intHour24) % 24;
		if (intHour24 < 0) intHour24 = 24 - intRemainder;
		else intHour24 = intRemainder;

	}
	
	// Roll over day / month / year
	if (intDay < 1) {
		intMonth--;
		if (intMonth < 1) {
			intMonth = 12;
			intYear--;
		}
		intDay = daysInMonth(intMonth, intYear);

	} else if (intDay > daysInMonth(intMonth, intYear)) {
		intMonth++;
		if (intMonth > 12) {
			intMonth = 1;
			intYear++;
		}
		intDay = 1;
	}



	// convert from 24 to 12 hour time.
	int intHour = intHour24;
	intHour %= 12;
	if (intHour == 0) intHour = 12;
	
	// Build the date strings.
	sprintf(strDate, "%02d-%s", intDay, arrMonths[intMonth - 1]);


	switch (ptrCluster->style) {
	
	case STYLE_DIGITAL:
	
		// Time
		font = fontXL;
		sprintf(strTime, "%d:%02d", intHour, intMinute);
		fontPrint(strTime, 130, 24, TEXT_ALIGN_RIGHT, true, false);

		// AM / PM
		font = fontS_White;
		sprintf(strTime, "%s", intHour24 < 12 ? "AM" : "PM");
		fontPrint(strTime, 132, 74, TEXT_ALIGN_LEFT, false, true);

		// Date
		font = colours[ptrCluster->colour].font;
		fontPrint(strDate, DISPLAY_WIDTH / 2, DISPLAY_HEIGHT - font.lineHeight, TEXT_ALIGN_CENTER, true, true);

		break;

	case STYLE_ANALOG:
		
		Display.media_SetAdd((word)(IMAGE_CLOCK >> 16), (word)IMAGE_CLOCK);
		Display.media_Image(0, 0);

		// Center
		x1	= DISPLAY_WIDTH / 2;
		y1	= DISPLAY_HEIGHT / 2;

		// Hour hand		
		rad	= (intHour * 60 + intMinute) * PI / 360;
		x2	= x1 + sin(rad) * HOUR_HAND_LENGTH;
		y2	= y1 - cos(rad) * HOUR_HAND_LENGTH;
		Display.gfx_Line(x1, y1, x2, y2, RED);

		// Minute hand
		rad	= intMinute * PI / 30;
		x2	= x1 + sin(rad) * MINUTE_HAND_LENGTH;
		y2	= y1 - cos(rad) * MINUTE_HAND_LENGTH;
		Display.gfx_Line(x1, y1, x2, y2, RED);

		break;
	}
}



void Gauge::drawHeading(void) {

	int	intFrame, intWidth;
	char strBuffer[5];

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

	if (!screenInitialised) {
		
		// Clear
		Display.gfx_Cls();
		screenInitialised = true;
	}
	
	switch (ptrCluster->style) {	
	case STYLE_DIGITAL:
		
		// N / E / S / W
		if ((ptrCluster->heading >= 337.5) || (ptrCluster->heading <  22.5)) intFrame = 0;	// N
		if ((ptrCluster->heading >=  22.5) && (ptrCluster->heading <  67.5)) intFrame = 1;	// NE
		if ((ptrCluster->heading >=  67.5) && (ptrCluster->heading < 112.5)) intFrame = 2;	// E
		if ((ptrCluster->heading >= 112.5) && (ptrCluster->heading < 157.5)) intFrame = 3;	// SE
		if ((ptrCluster->heading >= 157.5) && (ptrCluster->heading < 202.5)) intFrame = 4;	// S
		if ((ptrCluster->heading >= 202.5) && (ptrCluster->heading < 247.5)) intFrame = 5;	// SW
		if ((ptrCluster->heading >= 247.5) && (ptrCluster->heading < 292.5)) intFrame = 6;	// E
		if ((ptrCluster->heading >= 292.5) && (ptrCluster->heading < 337.5)) intFrame = 7;	// NW
		
		Display.media_SetAdd((word)(VIDEO_HEADING_DIGITAL >> 16), (word)VIDEO_HEADING_DIGITAL);
		Display.media_VideoFrame(16, 0, intFrame);
			
		// Angle
		font = colours[ptrCluster->colour].font;
		sprintf(strBuffer, "%d^", ptrCluster->heading);
		intWidth = fontWidth(strBuffer);
				
		// Print the angle
		fontPrint(strBuffer, DISPLAY_WIDTH / 2, DISPLAY_HEIGHT - font.lineHeight, TEXT_ALIGN_CENTER, true, true);

		break;

	case STYLE_ANALOG:
		
		Display.media_SetAdd((word)(VIDEO_HEADING_ANALOG >> 16), (word)VIDEO_HEADING_ANALOG);
		Display.media_VideoFrame(0, 0, ptrCluster->heading);
	
		break;
	}
}



void Gauge::drawGPSHome(void) {


	// To Do: Work out the angle and distance from the current lat/long to the home lat/long.
	
	DWORD dwrdAddress;
	char strDistance[8];
	char strBuffer[11];

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

	if (!screenInitialised) {
		
		// Clear
		Display.gfx_Cls();
		screenInitialised = true;
	}


	// Get decimal degrees.
	double	dblCurrentLatitude	= decimalDegrees(ptrCluster->latitude, ptrCluster->latitudeNS);
	double	dblCurrentLongitude	= decimalDegrees(ptrCluster->longitude, ptrCluster->longitudeEW);
	double	dblHomeLatitude		= decimalDegrees(ptrCluster->latitudeHome, ptrCluster->latitudeHomeNS);
	double	dblHomeLonitude		= decimalDegrees(ptrCluster->longitudeHome, ptrCluster->longitudeHomeEW);



	// Calculate the distance and bearing between Latitude/Longitude points.
	// See http://www.movable-type.co.uk/scripts/latlong.html

	static const int EARTH_RADIUS = 6371; // km

	// Let's get thing into Radians...
	double	lat1		= DegreesToRadians(dblCurrentLatitude);
	double	lat2		= DegreesToRadians(dblHomeLatitude);
	double	long1		= DegreesToRadians(dblCurrentLongitude);
	double	long2		= DegreesToRadians(dblHomeLonitude);
	double	deltaLat	= DegreesToRadians(dblHomeLatitude - dblCurrentLatitude);
	double	deltaLong	= DegreesToRadians(dblHomeLonitude - dblCurrentLongitude);

	// Distance
	double	a			= sin(deltaLat / 2) * sin(deltaLat / 2) + cos(lat1) * cos(lat2) * sin(deltaLong / 2) * sin(deltaLong / 2);
	double	c			= 2 * atan2(sqrt(a), sqrt(1 - a));
	float	dblDistance	= EARTH_RADIUS * c;

	// Bearing
	double	y			= sin(long2 - long1) * cos(lat2);
	double	x			= cos(lat1) * sin(lat2) - sin(lat1) * cos(lat2) * cos(long2 - long1);
	double	dblBearing	= RadiansToDegrees(atan2(y, x));

	// Adjust for our current heading.
	int intAngle		= (((int)dblBearing - (int)(ptrCluster->heading)) + 360) % 360;

	
	
	// Digital or Analog arrow?
	dwrdAddress = (ptrCluster->style == STYLE_DIGITAL) ? VIDEO_GPS_HOME_DIGITAL : VIDEO_GPS_HOME_ANALOG;

	// Arrow
	Display.media_SetAdd((word)(dwrdAddress >> 16), (word)dwrdAddress);
	Display.media_VideoFrame(24, 0, intAngle);
	
	// Distance
	font = (ptrCluster->style == STYLE_DIGITAL) ? colours[ptrCluster->colour].font : fontS_White;
	if (dblDistance < 1) {
		sprintf(strBuffer, "%d M", (int)(dblDistance * 1000));
	} else {
		ftoa(strDistance, dblDistance, 1);
		sprintf(strBuffer, "%s KM", strDistance);
	}
	fontPrint(strBuffer, DISPLAY_WIDTH / 2, DISPLAY_HEIGHT - font.lineHeight, TEXT_ALIGN_CENTER, true, true);

	/*
	Display.txt_FGcolour(WHITE);
	Display.gfx_MoveTo(0, 0);
	sprintf(strBuffer, "Angle: %d", intAngle);
	Display.putstr(strBuffer);
	*/
}


void Gauge::drawGPSLocation(void) {

	char strBuffer[51];
	int intDeg, intMin;
	double dblDeg, dblMin, dblSec;
	char strSec[6];

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;	

	if (!screenInitialised) {
		
		// Clear
		Display.gfx_Cls();
		Display.txt_FGcolour(WHITE);

		switch (ptrCluster->style) {		
		case STYLE_DIGITAL:
			// Print the label in the small font
			font = colours[ptrCluster->colour].font;
			fontPrint("LOCATION", DISPLAY_WIDTH / 2, DISPLAY_HEIGHT - font.lineHeight, TEXT_ALIGN_CENTER, false, false);
			
			font = fontL;
			break;

		case STYLE_ANALOG:
			Display.gfx_MoveTo(0, 20);
			Display.putstr("Latitude");

			Display.gfx_MoveTo(0, 74);
			Display.putstr("Longitude");

			font = fontM;
			break;
		}
		
		screenInitialised = true;
	}

	
	switch (ptrCluster->style) {
	
	case STYLE_DIGITAL:

		// Convert latitude from ddmm.mmmm to dd.dddddd
		ftoa(strBuffer, decimalDegrees(ptrCluster->latitude, ptrCluster->latitudeNS), 4);
		fontPrint(strBuffer, 0, 16, TEXT_ALIGN_LEFT, false, true);

		// Convert Longitude from ddmm.mmmm to dd.dddddd
		ftoa(strBuffer, decimalDegrees(ptrCluster->longitude, ptrCluster->longitudeEW), 4);
		fontPrint(strBuffer, 0, 64, TEXT_ALIGN_LEFT, false, true);

		break;

	case STYLE_ANALOG:

		// Convert Latitude from ddmm.mmmm to ddmm'ss.ss"
		intDeg	= ptrCluster->latitude / 100;
		intMin	= (int)(ptrCluster->latitude) % 100;
		dblSec	= (ptrCluster->latitude - (int)(ptrCluster->latitude)) * 60;
		ftoa(strSec, dblSec, 1);
		
		sprintf(strBuffer, "%d^%d'%s\"", intDeg, intMin, strSec);
		fontPrint(strBuffer, 0, 28, TEXT_ALIGN_LEFT, false, true);
		Display.gfx_RectangleFilled(cursorX, cursorY, DISPLAY_WIDTH - 1, cursorY + font.lineHeight - 1, BLACK);

		sprintf(strBuffer, "%c", ptrCluster->latitudeNS);
		Display.gfx_MoveTo(cursorX - 9, cursorY + 18);
		Display.putstr(strBuffer);

		// Convert Longitude from dddmm.mmmm to dddmm'ss.ss"
		intDeg	= ptrCluster->longitude / 100;
		intMin	= (int)(ptrCluster->longitude) % 100;
		dblSec	= (ptrCluster->longitude - (int)(ptrCluster->longitude)) * 60;
		
		ftoa(strSec, dblSec, 1);
		sprintf(strBuffer, "%d^%d'%s\"", intDeg, intMin, strSec);
		fontPrint(strBuffer, 0, 82, TEXT_ALIGN_LEFT, false, true);
		Display.gfx_RectangleFilled(cursorX, cursorY, DISPLAY_WIDTH - 1, cursorY + font.lineHeight - 1, BLACK);
		
		sprintf(strBuffer, "%c", ptrCluster->longitudeEW);
		Display.gfx_MoveTo(cursorX - 9, cursorY + 18);
		Display.putstr(strBuffer);

		break;
	
	}
}




void Gauge::drawSatellite(char *strLabel) {

	// Show the satellite fix animation.
	// The frame is determinied in the GaugeCluster::loop() function.
	
	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

	if (!screenInitialised) {
		
		// Clear
		Display.gfx_Cls();
					
		// Print the label			
		font = (ptrCluster->style == STYLE_DIGITAL) ? colours[ptrCluster->colour].font : fontS_White;
		fontPrint(strLabel, DISPLAY_WIDTH / 2, DISPLAY_HEIGHT - font.lineHeight, TEXT_ALIGN_CENTER, true, true);
		
		screenInitialised = true;
	}
			
	// Satellite Fix Animation
	Display.media_SetAdd((word)(VIDEO_SATELLITE_FIX >> 16), (word)VIDEO_SATELLITE_FIX);
	Display.media_VideoFrame(48, 24, ptrCluster->frame);
}



// Menus
// --------------------------------------------------------------------------------

void Gauge::exit(void) {

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

	offset = 0;
	cursor = 0;

	if (parent > 0) {
		
		mode = MODE_MENU;
		parent = menuItems[parent].parent;

		itemCount = 0;
		for (byte i = 0; i < ITEMCOUNT; i++) {
			if (menuItems[i].parent == parent) {
				items[itemCount++] = menuItems[i];
			}
		}

		drawMenu();
	}
	else setScreen(screen);
}



void Gauge::setMenu(int intParent) {
	
	mode				= MODE_MENU;
	parent				= intParent;
	offset				= 0;
	cursor				= 0;

	// Create an array of items only with this parent.
	// Note: Becuase this is an array of objects, they're passed by reference.
	// This means they'll retain their values when checked etc and exited the menu.
	itemCount = 0;
	for (byte i = 0; i < ITEMCOUNT; i++) {
		if ((menuItems[i].parent == intParent) && (itemCount < MAX_ITEMS_PER_MENU)) {
			items[itemCount++] = menuItems[i];
		}
	}

	drawMenu();	
}



void Gauge::drawMenu() {
	
	byte y;

	// Get a handle on this gauge's parent cluster.
	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;
	
	if (!screenInitialised) {

		// Clear
		Display.gfx_Cls();

		Display.txt_BGcolour(BLACK);
		Display.txt_FGcolour(colours[ptrCluster->colour].rgb565);
		Display.gfx_MoveTo(MENU_ITEM_PADDING_LEFT, MENU_TITLE_TOP);
		Display.putstr((parent == 0) ? (char *)"Main Menu" : menuItems[parent].name);

		screenInitialised = true;
	}

	// Up/Down arrows
	if (offset > 0) {
		Display.media_SetAdd((word)(IMAGE_CONTROL_ARROWUP >> 16), (word)IMAGE_CONTROL_ARROWUP);
		Display.media_Image(MENU_ARROW_LEFT, MENU_ARROW_TOP);
	} else {
		Display.gfx_RectangleFilled(MENU_ARROW_LEFT, MENU_ARROW_TOP, MENU_ARROW_LEFT + 11, MENU_ARROW_TOP + 5, BLACK);
	}
	if (offset < itemCount - ITEMS_PER_SCREEN) {
		Display.media_SetAdd((word)(IMAGE_CONTROL_ARROWDOWN >> 16), (word)IMAGE_CONTROL_ARROWDOWN);
		Display.media_Image(MENU_ARROW_LEFT, MENU_ARROW_BOTTOM);
	} else {
		Display.gfx_RectangleFilled(MENU_ARROW_LEFT, MENU_ARROW_BOTTOM, MENU_ARROW_LEFT + 11, MENU_ARROW_BOTTOM + 5, BLACK);
	}

	// Draw the menu items
	for (y = 0; y < min(ITEMS_PER_SCREEN, itemCount); y++) {
		drawItem(items[offset + y], y, (offset + y) == cursor);
	}
}



void Gauge::drawItem(MenuItemType objItem, byte y, boolean bHighlight) {

	DWORD dwrdImage;

	// Highlight the entire row or not.
    if (bHighlight) {
		Display.txt_BGcolour(WHITE);
		Display.gfx_RectangleFilled(0, y * MENU_ITEM_HEIGHT + MENU_ITEMS_TOP, DISPLAY_WIDTH - 1, y * MENU_ITEM_HEIGHT + MENU_ITEMS_TOP + MENU_ITEM_HEIGHT - 1, WHITE);
		Display.txt_FGcolour(BLACK);
    } else {
		Display.gfx_RectangleFilled(0, y * MENU_ITEM_HEIGHT + MENU_ITEMS_TOP, DISPLAY_WIDTH - 1, y * MENU_ITEM_HEIGHT + MENU_ITEMS_TOP + MENU_ITEM_HEIGHT - 1, BLACK);
		Display.txt_BGcolour(BLACK);
		Display.txt_FGcolour(WHITE);
	}

	
	// Print the text
	Display.gfx_MoveTo(MENU_ITEM_PADDING_LEFT, y * MENU_ITEM_HEIGHT + MENU_ITEM_BASELINE + MENU_ITEMS_TOP);
	Display.putstr(objItem.name);


	// Draw the icon / control
	switch (objItem.type) {
	
	case TYPE_ITEM:
		// Icon?
		if (objItem.icon != NONE) {
			dwrdImage = objItem.icon;
			if (bHighlight) dwrdImage += ICON_HIGHLIGHT_OFFSET;
			Display.media_SetAdd((word)(dwrdImage >> 16), (word)dwrdImage);
			Display.media_Image(MENU_ICON_PADDING_LEFT, y * MENU_ITEM_HEIGHT + MENU_ITEMS_TOP + MENU_ICON_PADDING_TOP);
		}
		break;
	
	case TYPE_CHECKBOX:

		if (getSetting(objItem)) {
			if (bHighlight)	dwrdImage = IMAGE_CONTROL_CHECKBOX_HIGHLIGHT_CHECKED;
			else			dwrdImage = IMAGE_CONTROL_CHECKBOX_CHECKED;
		} else {
			if (bHighlight)	dwrdImage = IMAGE_CONTROL_CHECKBOX_HIGHLIGHT;
			else			dwrdImage = IMAGE_CONTROL_CHECKBOX;
		}

		Display.media_SetAdd((word)(dwrdImage >> 16), (word)dwrdImage);
		Display.media_Image(MENU_CHECKBOX_PADDING_LEFT, y * MENU_ITEM_HEIGHT + MENU_ITEMS_TOP + MENU_CHECKBOX_PADDING_TOP);

		break;

	case TYPE_RADIOBUTTON:

		if (getSetting(objItem) == objItem.value) {
			if (bHighlight)	dwrdImage = IMAGE_CONTROL_RADIO_HIGHLIGHT_CHECKED;
			else			dwrdImage = IMAGE_CONTROL_RADIO_CHECKED;
		} else {
			if (bHighlight)	dwrdImage = IMAGE_CONTROL_RADIO_HIGHLIGHT;
			else			dwrdImage = IMAGE_CONTROL_RADIO;
		}

		Display.media_SetAdd((word)(dwrdImage >> 16), (word)dwrdImage);
		Display.media_Image(MENU_RADIO_PADDING_LEFT, y * MENU_ITEM_HEIGHT + MENU_ITEMS_TOP + MENU_RADIO_PADDING_TOP);

		break;
	}
}



word Gauge::getSetting(MenuItemType objItem) {

	word value;

	if (objItem.setting < CLUSTER_SETTINGS) {

		GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

		switch (objItem.setting) {
		case CLUSTER_STYLE:				value = ptrCluster->style;			break;
		case CLUSTER_TIMEZONE:			value = ptrCluster->timezone;		break;
		case CLUSTER_COLOUR:			value = ptrCluster->colour;			break;
		}
	} else {
		if (objItem.setting == GAUGE_CURRENT_SCREEN) value = screen;
		else value = (word)screens[objItem.setting - CLUSTER_SETTINGS];
	}

	return value;
}



void Gauge::setSetting(MenuItemType objItem, word value) {

	GaugeCluster * ptrCluster = (GaugeCluster *)cluster;

	if (objItem.setting < CLUSTER_SETTINGS) {

		switch (objItem.setting) {
		case CLUSTER_STYLE:				ptrCluster->style		= value;	break;
		case CLUSTER_TIMEZONE:			ptrCluster->timezone	= value;	break;
		case CLUSTER_COLOUR:			ptrCluster->colour		= value; screenInitialised = false; break;
		}
	} else {
		if (objItem.setting == GAUGE_CURRENT_SCREEN) screen = value;
		else screens[objItem.setting - CLUSTER_SETTINGS] = (boolean)value;
	}

	ptrCluster->saveSettings();

	ptrCluster->refreshAllGauges();
}



// --------------------------------------------------------------------------------



// +-------------------+
// | Class Constructor |
// +-------------------+
GaugeCluster::GaugeCluster() {

	gaugeCount		= 0;
	style			= STYLE_ANALOG;
	colour			= 3;	// Digital label colour (0 = White, 1 = Red, 2 = Orange, 3 = Yellow, 4 = Green, 5 = Blue)
	timezone		= 600;	// Minutes. 600 = +10 hours (Brisbane)

	rpm				= 0;
	rpmWas			= 0;
	knots			= 0;
	knotsWere		= 0;

	frame			= 0;
	frameWas		= VIDEO_SATELLITE_FIX_FRAMES - 1;
	frameTimer		= 0;

	speedoNeedsFix	= true;
	headingNeedsFix	= true;
	clockNeedsFix	= true;
}



void GaugeCluster::begin() {
	loadSettings();
	for (int g = 0; g < gaugeCount; g++) gauges[g].begin();	
}



void GaugeCluster::loop() {

	/*
	Button handling is done in a loop, rather than via events.
	This is to try and simulate the arduino loop().
	It also allows us to time how long a button was held for.
	
	The purpose of the loop is to:
	 - call .select()	when the button is pressed, and
	 - call .ok()		when the button is held
	
	This .loop() function should be called very regularly, ideally in your sketch"s loop().
	*/

	for (int g = 0; g < gaugeCount; g++) gauges[g].loop();
	
	// Update the "Last Drawn" values.
	fixWas				= fix;
	headingWas			= heading;
	knotsWere			= knots;
	rpmWas				= rpm;
	latitudeWas			= latitude;
	latitudeNSWas		= latitudeNS;
	longitudeWas		= longitude;
	longitudeEWWas		= longitudeEW;
	latitudeHomeWas		= latitudeHome;
	latitudeHomeNSWas	= latitudeHomeNS;
	longitudeHomeWas	= longitudeHome;
	longitudeHomeEWWas	= longitudeHomeEW;
	altitudeWas			= altitude;
    hourWas				= hour;
    minuteWas			= minute;
    secondsWere			= seconds;
    yearWas				= year;
    monthWas			= month;
    dayWas				= day;

	// what frame is the satellite animation on?
	frameWas = frame;
	if (millis() >= frameTimer) {
		frameTimer = millis() + VIDEO_SATELLITE_FIX_FRAME_DELAY;
		frame++;
		if (frame >= VIDEO_SATELLITE_FIX_FRAMES) frame = 0;
	}
}



void GaugeCluster::addGauge(Stream * virtualPort, int intBTN) {
	
	// Create the new gauge, and rememnber it"s index (for pointing it to the menu below).
	
	gaugeCount++;
	if (gaugeCount == 1) gauges = (Gauge *)malloc(sizeof(Gauge));
	else realloc(gauges, sizeof(Gauge) * gaugeCount);
	gauges[gaugeCount - 1] = Gauge(virtualPort, intBTN, this);
}



void GaugeCluster::refreshAllGauges() {
	for (int g = 0; g < gaugeCount; g++) gauges[g].refresh();
}



void GaugeCluster::loadSettings() {

	char strID[4];
	int address = SETTINGS_ADDRESS;

	// Try to read the ID string
	eeprom_read_block(strID, (void *)address, strlen(SETTINGS_ID_STRING));
	strID[strlen(SETTINGS_ID_STRING)] = 0; // terminate the newly read string
	address += strlen(SETTINGS_ID_STRING) + 1;

	// Echo the string that was read.
	Serial.print("Settings ID String (read from EEPROM): \"");
	Serial.print(strID);
	Serial.println("\".");

	if (strcmp(strID, SETTINGS_ID_STRING) == 0) {
		
		Serial.println("Valid ID found. Reading settings... ");

		// Gauge Cluster
		eeprom_read_block(&gaugeCount, (void *)address, sizeof(gaugeCount)); address += sizeof(gaugeCount);
		eeprom_read_block(&style, (void *)address, sizeof(style)); address += sizeof(style);
		eeprom_read_block(&colour, (void *)address, sizeof(colour)); address += sizeof(colour);
		eeprom_read_block(&timezone, (void *)address, sizeof(timezone)); address += sizeof(timezone);
		eeprom_read_block(&latitudeHome, (void *)address, sizeof(latitudeHome)); address += sizeof(latitudeHome);
		eeprom_read_block(&latitudeHomeNS, (void *)address, sizeof(latitudeHomeNS)); address += sizeof(latitudeHomeNS);
		eeprom_read_block(&longitudeHome, (void *)address, sizeof(longitudeHome)); address += sizeof(longitudeHome);
		eeprom_read_block(&longitudeHomeEW, (void *)address, sizeof(longitudeHomeEW)); address += sizeof(longitudeHomeEW);
		
		// Each Gauge
		for (int g = 0; g < gaugeCount; g++) {
			eeprom_read_block(&(gauges[g].screen), (void *)address, sizeof(gauges[g].screen)); address += sizeof(gauges[g].screen);
			for (int s = 0; s < SCREENS; s++) {
				eeprom_read_block(&(gauges[g].screens[s]), (void *)address, sizeof(gauges[g].screens[s])); address += sizeof(gauges[g].screens[s]);
			}
		}
	}
}



void GaugeCluster::saveSettings() {

	char strID[4];
	int address = SETTINGS_ADDRESS;

	// Try to write the ID string
	eeprom_write_block(SETTINGS_ID_STRING, (void *)address, strlen(SETTINGS_ID_STRING) + 1);
	address += strlen(SETTINGS_ID_STRING) + 1;
		
	// Gauge Cluster
	eeprom_write_block(&gaugeCount, (void *)address, sizeof(gaugeCount)); address += sizeof(gaugeCount);
	eeprom_write_block(&style, (void *)address, sizeof(style)); address += sizeof(style);
	eeprom_write_block(&colour, (void *)address, sizeof(colour)); address += sizeof(colour);
	eeprom_write_block(&timezone, (void *)address, sizeof(timezone)); address += sizeof(timezone);
	eeprom_write_block(&latitudeHome, (void *)address, sizeof(latitudeHome)); address += sizeof(latitudeHome);
	eeprom_write_block(&latitudeHomeNS, (void *)address, sizeof(latitudeHomeNS)); address += sizeof(latitudeHomeNS);
	eeprom_write_block(&longitudeHome, (void *)address, sizeof(longitudeHome)); address += sizeof(longitudeHome);
	eeprom_write_block(&longitudeHomeEW, (void *)address, sizeof(longitudeHomeEW)); address += sizeof(longitudeHomeEW);
	
	// Each Gauge
	for (int g = 0; g < gaugeCount; g++) {
		eeprom_write_block(&(gauges[g].screen), (void *)address, sizeof(gauges[g].screen)); address += sizeof(gauges[g].screen);
		for (int s = 0; s < SCREENS; s++) {
			eeprom_write_block(&(gauges[g].screens[s]), (void *)address, sizeof(gauges[g].screens[s])); address += sizeof(gauges[g].screens[s]);
		}
	}
}



// --------------------------------------------------------------------------------




// +----------------+
// | Misc Functions |
// +----------------+

byte daysInMonth(byte bytMonth, int intYear) {
	
	static const byte arrDaysInMonth[]	= { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

	byte bytDaysInMonth = 0;

	// Days per month
	bytDaysInMonth = arrDaysInMonth[bytMonth - 1];
	if ((bytMonth == 2) && (intYear % 4 == 0)) bytDaysInMonth++;

	return bytDaysInMonth;
}



char *ftoa(char *a, double f, int precision) {
 
	long p[] = { 0, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000 };

	char *ret = a;
	long heiltal = (long)f;
	itoa(heiltal, a, 10);
	while (*a != '\0') a++;
	*a++ = '.';
	long desimal = abs((long)((f - heiltal) * p[precision]));
	itoa(desimal, a, 10);
	return ret;
}



double decimalDegrees(double dblDDDMMpointMMMM, char NSEW) {
	int intDeg = (dblDDDMMpointMMMM / 100);
	double dblMin = fmod(dblDDDMMpointMMMM, 100);
	double dblDeg = (dblMin / 60) + intDeg;
	if ((NSEW == 'S') || (NSEW == 'W')) dblDeg *= -1;
	return dblDeg;
}



double DegreesToRadians(double dblDegrees) { return dblDegrees * (PI / 180); }
double RadiansToDegrees(double dblRadians) { return dblRadians * (180 / PI); }